/*	scheduler.c

	Tuple scheduler and register allocator.
*/



#undef DEBUG
#undef DEBUG_BEFOREBB
#undef DEBUG_ENDBB
#undef DEBUG_SEARCH
#undef DEBUG_PERF
#undef DEBUG_TUPLIST
#undef DEBUG_m2rmode
#undef DEBUG_SYMNAME
#undef DEBUG_SCHED
#undef DEBUG_CODEGEN

#undef NEWSTUFF

#include "stdpccts.h"
#include "swartypes.h"
#include "tuple.h"
#include "oputils.h"
#include "scheduler.h"
#include "spool.h"
#include "showir.h"

extern char     *funcname;      /* current function name */


static tuple	safetup[MAXTUP];	/* tuple buffer */
tuple	*tup = &(safetup[1]);	/* tuple buffer */
int	tupsp;			/* tuple stack pointer */

static int	sched[MAXTUP];		/* tuple schedule */
static int	schedsp = 0;

#define NOTFOUND -1
static int	sbest[MAXTUP];		/* best schedule so far */
static int	sbestval = NOTFOUND;	/* value of best schedule */
static int	schedval = 0;		/* value of current schedule */

static int	schedtim;		/* time to stop scheduler */


void
begin_bb(void)
{
	/* Reinitialize the tuple stack pointer
	   and fieldcount for the next block
	*/
	tupsp = 0;
	fieldcount = 0;
}

static void
incref(int i)
{
	/* Increment refs of live tuples */
	/* Prerequisite tuples are live */

	if (i >= MAXTUP)
		bug("tuple number is greater than MAXTUP in tuple.h");

	if ((i >= 0) &&
	    (++(tup[i].refs) == 1)) {
		if (tup[i].arg[0] != -1) incref(tup[i].arg[0]);
		if (tup[i].arg[1] != -1) incref(tup[i].arg[1]);
	}
}

static int
before_bb(int here,
int there,
int init)
{
	/* Returns 1 if "here" is a prerequiste for "there".
	   Returns 0 otherwise.
	*/
	static int cache[MAXTUP][MAXTUP];
	int rval;

	#ifdef DEBUG_BEFOREBB
		info(0, "Start before_bb");
	#endif

	if (init) {
		#ifdef DEBUG_BEFOREBB
			info(0, "before_bb():start init loop");
		#endif

		for (here=0; here<tupsp; ++here)
			for (there=0; there<tupsp; ++there)
				cache[here][there] = -1;

		#ifdef DEBUG_BEFOREBB
			info(0, "before_bb():end init loop");
		#endif

		return 0;
	}

	if (cache[here][there] != -1) return(cache[here][there]);

	if ((there < 0) ||
	    (tup[there].refs < 1)) {
		cache[here][there] = 0;
		return(0);
	}

	#ifdef DEBUG_BEFOREBB
		info(0, "before_bb: a");
	#endif
	if (here == there) {
		cache[here][there] = 1;
		return(1);
	}

	#ifdef DEBUG_BEFOREBB
		info(0, "before_bb: b");
	#endif
	switch (tup[there].op) {
	case ADD:
	case AND:
	case ANDN:
	case AVG:
	case DIV:
	case EQ:
	case EQ_C:
	case GE:
	case GT:
	case GT_C:
	case MOD:
	case MUL:
	case MULH:
	case OR:
	case SHL:
	case SHR:
	case SUB:
	case XOR:
	case PACK:
	case UNPACKL:
	case UNPACKH:
	case MIXEVEN:
	case MIXODD:
	case PERM:
	case MAX:
	case MIN:
	case RCP1:
	case RCP2:
	case STORER:
		/* Binary op */
		#ifdef DEBUG_BEFOREBB
			info(0, "before_bb: c");
		#endif
		rval = (before_bb(here, tup[there].arg[0], 0) ||
		        before_bb(here, tup[there].arg[1], 0));
		cache[here][there] = rval;
		return rval;
	case I2F:
	case F2I:
	case LNOT:
	case NEG:
	case NOT:
	case LEA:
	case LOADR:
	case STORE:
	case RCP:
		/* Unary op */
		#ifdef DEBUG_BEFOREBB
			info(0, "before_bb: d");
		#endif
		tup[there].arg[1] = -1;
		rval = before_bb(here, tup[there].arg[0], 0);
		cache[here][there] = rval;
		return rval;

	default:
		{
			char buf[64];
			snprintf(buf,
				64,
				"before_bb op=%s (this cannot happen?)",
				opname(tup[there].op));
			bug(buf);
		}
		/* Fall through... */
	case LOAD:
	case NUM:
		tup[there].arg[0] = -1;
		tup[there].arg[1] = -1;
		break;
	}

	#ifdef DEBUG_BEFOREBB
		info(0, "before_bb: e");
	#endif
	cache[here][there] = 0;
	return(0);
}

static int
pickreg(register int pos)
{
	/* Find and use a free register.
	   Returns 0 if a register was found, 1 if not.
	*/

	register int i, j, k;
	int regbusy[MAXSPILLS+REGSAVAIL];	/* -1 if not in use */
	int regwho[MAXSPILLS+REGSAVAIL];	/* Tuple using this register */
	register int dontuse = MAXSPILLS+REGSAVAIL;

	/* Mark all registers as not busy and never used */
	for (i=0; i<MAXSPILLS+REGSAVAIL; ++i) {
		regbusy[i] = -1;
		regwho[i] = -1;
	}

	/* If there is a 1st argument to this tuple, don't use its register? */
	if (tup[ sched[pos] ].arg[0] != -1) {
		dontuse = tup[ tup[ sched[pos] ].arg[0] ].reg;
	}

	/* Find/Store information about previous tuples in the list */
	for (i=0; i<=pos; ++i) {
		/* Mark busy registers with the time (i.e. sched[]) slot in
		   which they are used */
		if ((tup[ sched[i] ].reg != -1) &&
		    (tup[ sched[i] ].trefs > 0)) {
			regbusy[ tup[ sched[i] ].reg ] = i;
		}

		/* Mark when each argument was last used in regwho */
		if ((tup[ sched[i] ].arg[0] != -1) &&
		    (regwho[ tup[ tup[ sched[i] ].arg[0] ].reg ] == -1)) {
			regwho[ tup[ tup[ sched[i] ].arg[0] ].reg ] = i;
		}
		if ((tup[ sched[i] ].arg[1] != -1) &&
		    (regwho[ tup[ tup[ sched[i] ].arg[1] ].reg ] == -1)) {
			regwho[ tup[ tup[ sched[i] ].arg[1] ].reg ] = i;
		}
	}

	/* Set j to the least recently used free register? */
	/* Uses lowest numbered when all are free */
	k = pos;
	j = numspills+REGSAVAIL;
	for (i=0; i<numspills+REGSAVAIL; ++i) {
		if (regbusy[i] == -1) {
			if ((regwho[i] < k) &&
			    (dontuse != i)) {
				j = i;
				k = regwho[i];
			}
		}
	}

	/* If j is a valid register or pseudoregister, use it */
	if (j < numspills+REGSAVAIL) {
		switch (tup[ sched[pos] ].op) {
		case LOAD:
		case NUM:
			/* Use j for this tuple */
			tup[ sched[pos] ].reg = j;
			return(0);
		default:
			/* Give j to the first argument and use its
			   register instead */
			tup[ sched[pos] ].reg =
			  tup[ tup[ sched[pos] ].arg[0] ].reg;
			tup[ tup[ sched[pos] ].arg[0] ].reg = j;
			tup[ sched[pos] ].antidep = j;
			return(0);
		}
	}

	/* If j is not valid, no registers are available */
	return(1);
}

static int
argregs(register int n)
{
	/* Allocate register(s?) for the tuple corresponding
	     to position n in the candidate schedule.
	   Returns 0 if register allocation was okay.
	   Returns 1 otherwise.
	*/

	register int s = sched[n];

	/* Temporary reference count.  We'll decrement this as we allocate
	   time slots to referencing tuples.
	*/
	tup[s].trefs = tup[s].refs;

	/* Clear previous register assignment info */
	tup[s].oreg = -1;
	tup[s].reg = -1;

	if ((tup[s].op == STORE) || (tup[s].op == STORER)) {
		/* stores do not need registers (to hold the result?) */
		--(tup[ tup[s].arg[0] ].trefs);
	} else {
		tup[s].antidep = -1;

		if (tup[s].arg[0] != -1) {
			/* Plan to use arg0's register... */
			tup[s].reg = tup[ tup[s].arg[0] ].reg;

			if (--(tup[ tup[s].arg[0] ].trefs) > 0) {
				/* ...but if it is referenced by later tuples,
				   pick a new register */
				if (pickreg(n)) return(1);
			}
		} else {
			if (pickreg(n)) return(1);
		}

		if (tup[s].arg[1] != -1) {
			/* free register not overwritten */
			--(tup[ tup[s].arg[1] ].trefs);
		}
	}
	tup[s].oreg = tup[s].reg;

	return(0);
}

static void
unregs(register int n)
{
	register int s = sched[n];

	if ((tup[s].op == STORE) || (tup[s].op == STORER)) {
		/* stores do not need registers (to hold the result?) */
		++(tup[ tup[s].arg[0] ].trefs);
	} else {
		if (tup[s].arg[1] != -1) {
			/* free register not overwritten */
			++(tup[ tup[s].arg[1] ].trefs);
		}

		if (tup[s].arg[0] != -1) {
			++(tup[ tup[s].arg[0] ].trefs);

			if (tup[s].antidep != -1) {
				tup[ tup[s].arg[0] ].reg = tup[s].oreg;
			}
		}
	}
}

static int perm_count = 0;

static int
sched_perm(register int pos)
{
	/* Recursively build a schedule.
	   Returns 0 if this schedule is not better than the current best.
	   Otherwise, if we have a comlete schedule(?), returns its cost.
	   If the schedule is not complete:
		return 1 if timed-out during subschedule
		return 1 if sucessful building subschedule
		return 0 if failed building subschedule
	*/

	register int i;

	#ifdef DEBUG_SCHED
		fprintf(stderr, "Start sched_perm(%d):\n", pos);
		fprintf(stderr, " sbestval=%d, schedval=%d schedsp=%d, "
				"perm_count=%d, schedtim=%d\n",
				sbestval, schedval, schedsp, perm_count,
				schedtim);
	#endif

	/* If we already have a best schedule,
	   and this one isn't better, return 0. */
	if ((sbestval != NOTFOUND) &&
	    (schedval >= sbestval)) return(0);

	/* If a complete schedule, return its cost */
	if (pos >= schedsp) {
		/* Have a complete schedule */
		return(sched_eval());
	}

/* HEREHERE - What is this? */
	/* Timeout if needed... */
	if (--perm_count < 0) {
		perm_count = 2 * schedsp;
		if (time(0) >= schedtim) {
			return(1);
		}
	}

	/* Hack to improve scheduling...
	   if next instruction depends on previous,
	   swap this with next op in perm order
	*/
	if ((pos > 1) &&
	    ((pos + 1) < schedsp) &&
	    ((tup[sched[pos]].arg[0] == sched[pos-1]) ||
	     (tup[sched[pos]].arg[1] == sched[pos-1]))) {
		i = sched[pos];
		sched[pos] = sched[pos+1];
		sched[pos+1] = i;
	}

	/* Only a partial schedule, continue building */
	for (i=pos; i<schedsp; ++i) {
		if (sched_ok(i, pos)) {
			register int r = -1;
			register int j;
			register int k;

			j = sched[i];
			sched[i] = sched[pos];
			sched[pos] = j;

			if ((tup[ sched[pos] ].arg[1] != -1) &&
			    (!ordered(tup[ sched[pos] ].op)) &&
			    (tup[ tup[ sched[pos] ].arg[0] ].trefs >
			     tup[ tup[ sched[pos] ].arg[1] ].trefs)) {
				/* normalize arg order to make the first
				   arg the one with the lowest reference
				   count, hence least likely to need a
				   new register when this clobbers it
				*/
				register int t;
				t = tup[ sched[pos] ].arg[0];
				tup[ sched[pos] ].arg[0] =
					tup[ sched[pos] ].arg[1];
				tup[ sched[pos] ].arg[1] = t;
			}

			/* if we allocated registers ok, recurse */
			if (!argregs(pos)) {
				k = sched_cost(pos);
				schedval += k;

				/* If we recurse successfully, return success */
				if (sched_perm(pos + 1)) return(1);

				/* Otherwise, backtrack */
				schedval -= k;
			}

			/* If we are here, we must undo this attempt */

			/* undo the register allocation we did */
			unregs(pos);

			sched[pos] = sched[i];
			sched[i] = j;
		}
	}

	/* If we got here, we failed */
	return(0);
}

static char *
symname(sym *s,
int offset)
{
	static char symnamebuf[256];

	#ifdef DEBUG_SYMNAME
		fflush(Cout);
		fprintf(stderr,
			"symname(): s=%p s->text=%s offset=%d\n",
			s,
			(s)? ((s->text)? s->text : "(null)") : "-",
			offset );
		fflush(stderr);
	#endif

	snprintf(&(symnamebuf[0]),
		256,
		"*(((p%d_t *) %s%s) + %d)",
		bitsperfrag(),
		((s->scope == 1) ? "" : "&"),
		s->text,
		tup[offset].immed);
	return(&(symnamebuf[0]));
}

static char *
show_suf2(int i)
{
	if (i < 1) return("q");
	switch (i) {
	case 1:		return("1");
	case 2:		return("2");
	case 4:		return("4");
	case 8:		return("b");
	case 16:	return("w");
	case 32:	return("d");
	}
	return("q");
}

static char *
show_suf(int i)
{
	return(show_suf2(tup[i].type.bits));
}

static char *
ia32regname(int regnum)
{
	static char buf[7];
	#ifdef DEBUG_CODEGEN
		int rval = 
	#endif

	snprintf ( buf, 6, "reg%d", regnum );
	#ifdef DEBUG_CODEGEN
	    fprintf(Cout,
		"\n/* ia32regname(%d) buf is %s; return value is %d*/\n",
		regnum,
		buf,
		rval);
	#endif

	return buf;
}

static void
ia32op(register int s)
{
	register int r0 = -1;
	register int r1 = -1;
	char buf[256];		/* This is much larger than it needs to be */

	/* Get first arg, with possible rename */
	if (tup[s].arg[0] != -1) {
		r0 = tup[ tup[s].arg[0] ].reg;

		if (tup[s].antidep != -1) {
			/* Arg was renamed; output a move */
			r0 = tup[s].reg;
			Ctab();

			/* Note ia32regname returns a pointer to a
			   static buffer, so we need to use its value
			   before calling it a second time.  Thus the
			   separate fprintf()s.
			*/
			fprintf(Cout, "movl_r2r(%s,", ia32regname(r0));
			fprintf(Cout,
				" %s);\n",
				ia32regname(tup[s].antidep));
			#ifdef DEBUG_CODEGEN
				fprintf(Cout,"/* %d */\n",__LINE__);
			#endif
			tup[ tup[s].arg[0] ].reg = tup[s].antidep;
		}
	}

	/* Get second arg */
	if (tup[s].arg[1] != -1) {
		r1 = tup[ tup[s].arg[1] ].reg;
	}

	/* Where does result go? */
	tup[s].reg = r0;

	Ctab();
	switch (tup[s].op) {
	case F2I:
		bug("float->int cast not available for this target");
		break;
	case I2F:
		bug("int->float cast not available for this target");
		break;
	case RCP:
		bug("float reciprocal not available for this target");
		break;
	case RCP1:
		bug("float reciprocal not available for this target");
		break;
	case RCP2:
		bug("float reciprocal not available for this target");
		break;
	case ADD:
		if (tup[s].type.attr & TYP_FLOAT) {
			bug("float ADD not available for this target");
		} else {
			sprintf(buf, "addl");
		}
		break;
	case SUB:
		if (tup[s].type.attr & TYP_FLOAT) {
			bug("float SUB not available for this target");
		} else {
			sprintf(buf, "subl");
		}
		break;
	case MUL:
		if (tup[s].type.attr & TYP_FLOAT) {
			bug("float MUL not available for this target");
		} else {
			sprintf(buf, "mulll");
		}
		break;
	case MULH:
		if (tup[s].type.attr & TYP_FLOAT) {
			bug("float MULH not available for this target");
		} else if (tup[s].type.attr & TYP_UNSIGN) {
			sprintf(buf, "mulhul");
		} else {
			sprintf(buf, "mulhl");
		}
		break;
	case DIV:
		snprintf(buf,
			256,
			"div%sl",
			((tup[s].type.attr & TYP_UNSIGN) ? "u" : ""));
		break;
	case MOD:
		snprintf(buf,
			256,
			"mod%sl",
			((tup[s].type.attr & TYP_UNSIGN) ? "u" : ""));
		break;
	case AVG:
		bug("AVG instruction not available for this target");
		break;
	case MAX:
		bug("MAX instruction not available on this target");
		break;
	case MIN:
		bug("MIN instruction not available for this target");
		break;
	case AND:
		sprintf(buf, "andl");
		break;
	case ANDN:
		sprintf(buf, "andnl");
		break;
	case OR:
		sprintf(buf, "orl");
		break;
	case XOR:
		sprintf(buf, "xorl");
		break;
	case EQ:
		bug("EQ instruction not available for this target");
		break;
	case EQ_C: /* Compare for EQ, with C-like result */
		sprintf(buf, "cmpeql");
		break;

	case GT:
		bug("GT instruction not available for this target");
		break;
	case GT_C: /* Compare for GT, with C-like result */
		if (tup[s].type.attr & TYP_UNSIGN) {
			sprintf(buf, "cmpgtul");
		} else {
			sprintf(buf, "cmpgtl");
		}
		break;

	case GE:
		bug("GE instruction not available for this target");
		break;

	case SHL:
		sprintf(buf, "slll");
		break;

	case SHR:
		sprintf(buf, "sr%cl",
			(tup[s].type.attr & TYP_UNSIGN)?
			    'l' :
			    (tup[s].type.attr == TYP_NULL)? 'l' : 'a');
		break;

	case PACK:
		bug("PACK instruction not available for this target");
		break;

	case UNPACKL:
		bug("PUNPCKL instruction not available for this target");
		break;
	case UNPACKH:
		bug("PUNPCKH instruction not available for this target");
		break;

	case MIXEVEN:
		bug("MIXEVEN instruction not available for this target");
		break;

	case MIXODD:
		bug("MIXODD instruction not available for this target");
		break;

	case PERM:
		bug("PERM instruction not available for this target");
		break;

	case LEA:
		snprintf(buf, 256, "addl");
		break;

	case LOADR:
		snprintf(buf, 256, "movl");
		break;

	default:
		{
			char buf[64];
			snprintf(buf,
				64,
				"ia32op op=%s (this cannot happen?)",
				opname(tup[s].op));
			bug(buf);
		}
		return;
	}

	/* Output the arguments */
	#ifdef DEBUG_m2rmode
		fprintf(Cout, "/* m2rmode is %d */", tup[s].m2rmode);
	#endif
	if (tup[s].m2rmode != -1) {
		if (tup[ tup[s].m2rmode ].op == LOAD) {
			fprintf(Cout, "%s_m2r(%s, %s);",
				buf,
				symname(tup[ tup[s].m2rmode ].symbol,
					tup[ tup[s].m2rmode ].offset),
				ia32regname(r0));
			#ifdef DEBUG_CODEGEN
				fprintf(Cout,"/* %d */\n",__LINE__);
			#endif
		} else {
			/* an immediate from the constant pool */
			switch (tup[s].op) {
			case SHL:
			case SHR:
			case PERM:
				/* do not need constant pool for 8-bit
				   immediates, but this only works for shifts
				   and adds a byte to the opcode, so it is not
				   as clear a win as one might hope...  anyway,
				   use it if we can
				*/
				/* This is touchy about the return from
				   ia32regname().  I'm not sure why
				   because it seems to me that the
				   fprintf()s should be combinable.
				   Maybe a segment pointer gets
				   clobbered?
				*/
				fprintf(Cout, "%s_i2r(%ld, ",
					buf,
					(tup[tup[s].m2rmode].immed.d[0]
					 & 0xffLL));
				fprintf(Cout, "%s);", ia32regname(r0));
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,
						"/* %d */\n",
						__LINE__);
				#endif
				break;
			default:
				/* load frag-sized immediates via the constant
				   pool
				*/
				cpoolenter(tup[ tup[s].m2rmode ].immed);

				fprintf(Cout, "%s_m2r(%s, %s);",
					buf,
					cpoolname(tup[ tup[s].m2rmode ].immed),
					ia32regname(r0));
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,"/* %d */\n",__LINE__);
				#endif
			}
		}
	} else {
		if (tup[s].op == LEA) {
			/* Generate the address */
			#ifdef DEBUG_CODEGEN
				fprintf(Cout,"/* %d */\n",__LINE__);
			#endif
			fprintf(Cout,
				"%s_m2r(((p32_t)((unsigned int)%s)), %s);",
				buf,
				tup[s].symbol->text,
				ia32regname(r0));
		} else if (tup[s].op == LOADR) {
			/* Generate the load using address in register */
			#ifdef DEBUG_CODEGEN
				fprintf(Cout,"/* %d */\n",__LINE__);
			#endif
			/* We can get away without splitting this for the two
			   separate ia32regname() calls because they have the
			   value anyway. */
			fprintf(Cout,
				"%s_x2r(%s, %s);",
				 buf,
				 ia32regname(r0),
				 ia32regname(r0));
		} else {
			/* Note ia32regname returns a pointer to a
			   static buffer, so we need to use its value
			   before calling it a second time.  Thus the
			   separate fprintf()s.
			*/
			#ifdef DEBUG_CODEGEN
				fprintf(Cout,"/* %d */\n",__LINE__);
			#endif
			fprintf(Cout, "%s_r2r(%s,", buf, ia32regname(r1));
			fprintf(Cout, " %s);", ia32regname(r0));
		}
	}
	fprintf(Cout, "\n");
}

static void
mmxop(register int s)
{
	register int r0 = -1;
	register int r1 = -1;
	char buf[256];		/* This is much larger than it needs to be */

	/* Get first arg, with possible rename */
	if (tup[s].arg[0] != -1) {
		r0 = tup[ tup[s].arg[0] ].reg;

		if (tup[s].antidep != -1) {
			/* Arg was renamed; output a move */
			r0 = tup[s].reg;
			Ctab();

			if ( (r0 < REGSAVAIL) &&
			     (tup[s].antidep < REGSAVAIL) ) {
				/* Neither spilled */
				fprintf(Cout,
					"movq_r2r(mm%d, mm%d);\n",
					r0,
					tup[s].antidep);
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,
						"/* %d */\n",
						__LINE__);
				#endif
			} else if ( (r0 >= REGSAVAIL) &&
			     (tup[s].antidep < REGSAVAIL) ) {
				/* Source spilled */
				fprintf(Cout,
					"movq_m2r(%s, mm%d);\n",
					spoolname(r0),
					tup[s].antidep);
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,
						"/* %d */\n",
						__LINE__);
				#endif
			} else if ( (r0 < REGSAVAIL) &&
			     (tup[s].antidep >= REGSAVAIL) ) {
				/* Destination spilled */
				fprintf(Cout,
					"movq_r2m(mm%d, %s);\n",
					r0,
					spoolname(tup[s].antidep));
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,
						"/* %d */\n",
						__LINE__);
				#endif
			} else if ( (r0 >= REGSAVAIL) &&
			     (tup[s].antidep >= REGSAVAIL) ) {
				/* Source and destination spilled */
				/* Save mm0 */
				fprintf(Cout,
					"movq_r2m(mm0, %s);\n",
					spoolname(REGSAVAIL-1));
				/* Move source into mm0 */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, mm0);\n",
					spoolname(r0));
				/* Move source into destination */
				Ctab();
				fprintf(Cout,
					"movq_r2m(mm0, %s);\n",
					spoolname(tup[s].antidep));
				/* Restore mm0 */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, mm0);\n",
					spoolname(REGSAVAIL-1));
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,
						"/* %d */\n",
						__LINE__);
				#endif
			}
			tup[ tup[s].arg[0] ].reg = tup[s].antidep;
		}
	}

	/* Get second arg */
	if (tup[s].arg[1] != -1) {
		r1 = tup[ tup[s].arg[1] ].reg;
	}

	/* Where does result go? */
	tup[s].reg = r0;

	Ctab();
	switch (tup[s].op) {
	case F2I:
		if ((optcpu & CPU_athlon) || (optcpu & CPU_3DNow)) {
			sprintf(buf, "pf2i%s", show_suf(s));
			r1 = r0;
		} else {
			bug("float->int cast available only for (ext'd)3DNow!");
		}
		break;
	case I2F:
		if ((optcpu & CPU_athlon) || (optcpu & CPU_3DNow)) {
			sprintf(buf, "pi2f%s", show_suf(s));
			r1 = r0;
		} else {
			bug("int->float cast available only for (ext'd)3DNow!");
		}
		break;
	case RCP:
		if (optcpu & CPU_3DNow) {
			sprintf(buf, "pfrcp");
			r1 = r0;
		} else {
			bug("float reciprocal available only for 3DNow!");
		}
		break;
	case RCP1:
		if (optcpu & CPU_3DNow) {
			sprintf(buf, "pfrcpit1");
		} else {
			bug("float reciprocal available only for 3DNow!");
		}
		break;
	case RCP2:
		if (optcpu & CPU_3DNow) {
			sprintf(buf, "pfrcpit2");
		} else {
			bug("float reciprocal available only for 3DNow!");
		}
		break;

	case ADD:
		if (tup[s].type.attr & TYP_FLOAT) {
			if (optcpu & CPU_3DNow) {
				sprintf(buf, "pfadd");
			} else {
				bug("float ADD available only for 3DNow!");
			}
		} else {
			if (optcpu & CPU_MMX) {
				snprintf(buf, 256, "padd%s%s%s",
					 ((tup[s].type.attr & TYP_SAT) &&
					  (tup[s].type.attr & TYP_UNSIGN)
					 )?  "u": "",
					 (tup[s].type.attr & TYP_SAT)? "s": "",
					 show_suf(s));
			} else {
				bug("ADD instruction not available for this "
				    "target");
			}
		}
		break;
	case SUB:
		if (tup[s].type.attr & TYP_FLOAT) {
			if (optcpu & CPU_3DNow) {
				sprintf(buf, "pfsub");
			} else {
				bug("float SUB available only for 3DNow!");
			}
		} else {
			if (optcpu & CPU_MMX) {
				snprintf(buf, 256, "psub%s%s%s",
					 ((tup[s].type.attr & TYP_SAT) &&
					  (tup[s].type.attr & TYP_UNSIGN)
					 )?  "u": "",
					 (tup[s].type.attr & TYP_SAT)? "s": "",
					 show_suf(s));
			} else {
				bug("SUB instruction not available for this "
				    "target");
			}
		}
		break;
	case MUL:
		if (tup[s].type.attr & TYP_FLOAT) {
			if (optcpu & CPU_3DNow) {
				sprintf(buf, "pfmul");
			} else {
				bug("float MUL available only for 3DNow!");
			}
		} else {
			if (optcpu & CPU_MMX)
				snprintf(buf, 256, "pmull%s", show_suf(s));
			else
				bug("MUL instruction not available for this "
				    "target");
		}
		break;
	case MULH:
		if (tup[s].type.attr & TYP_FLOAT) {
			bug("float MULH instruction not available for this "
			    "target");
		} else if (tup[s].type.attr & TYP_UNSIGN) {
			if ((optcpu & CPU_MMX) || (optcpu & CPU_athlon))
				snprintf(buf, 256, "pmulhu%s", show_suf(s));
			else
				bug("MULH instruction not available for this "
				    "target");
		} else {
			if (optcpu & CPU_MMX)
				snprintf(buf, 256, "pmulh%s", show_suf(s));
			else
				bug("MULH instruction not available for this "
				    "target");
		}
		break;
	case DIV:
		if (optcpu & CPU_MMX)
			snprintf(buf,
				256,
				"pdiv%s%s",
				((tup[s].type.attr & TYP_UNSIGN) ? "u" : ""),
				show_suf(s));
		else
			bug("DIV instruction not available for this target");
		break;
	case MOD:
		if (optcpu & CPU_MMX)
			snprintf(buf,
				256,
				"pmod%s%s",
				((tup[s].type.attr & TYP_UNSIGN) ? "u" : ""),
				show_suf(s));
		else
			bug("MOD instruction not available for this target");
		break;
	case AVG:
		if (optcpu & CPU_athlon) {
			sprintf(buf, "pavg%s", show_suf(s));
		} else if ((optcpu & CPU_3DNow) || (optcpu & CPU_XMMX)) {
			if (bitsperfield(tup[s].type.bits) != 8)
				bug("tried to generate AVG instruction for "
				    "other than 8 bit fields (this cannot "
				    "happen?)");
			sprintf(buf, "%s%s%s%s",
				((optcpu & CPU_XMMX) ? "pave" : "pavg"),
				((tup[s].type.attr & TYP_UNSIGN) ? "u" : ""),
				((tup[s].type.attr & TYP_SAT) ? "s" : ""),
				show_suf(s));
		} else {
			bug("AVG instruction not available for this target");
		}
		break;

	case MAX:
		if (tup[s].type.attr & TYP_FLOAT) {
			if (optcpu & CPU_3DNow) {
				sprintf(buf, "pfmax");
			} else {
				bug("float MAX available only for 3DNow!");
			}
		} else {
			if (optcpu & CPU_XMMX) {
				sprintf(buf, "pmag%s", show_suf(s));
			} else if (optcpu & CPU_athlon) {
				if (!(tup[s].type.attr & TYP_UNSIGN) &&
				     (bitsperfield(tup[s].type.bits)==16)) {
					sprintf(buf, "pmaxsw");
				} else if ((tup[s].type.attr & TYP_UNSIGN) &&
					  (bitsperfield(tup[s].type.bits)==8)) {
						sprintf(buf, "pmaxub");
				}
			} else {
				bug("MAX instruction not available on this "
				    "target");
			}
		}
		break;
	case MIN:
		if (tup[s].type.attr & TYP_FLOAT) {
			if (optcpu & CPU_3DNow) {
				sprintf(buf, "pfmin");
			} else {
				bug("float MIN available only for 3DNow!");
			}
		} else {
			if (optcpu & CPU_athlon) {
				if (!(tup[s].type.attr & TYP_UNSIGN) &&
				     (bitsperfield(tup[s].type.bits)==16)) {
					sprintf(buf, "pminsw");
				} else if ((tup[s].type.attr & TYP_UNSIGN) &&
					  (bitsperfield(tup[s].type.bits)==8)) {
						sprintf(buf, "pminub");
				}
			} else {
				bug("MIN instruction not available for this "
				    "target");
			}
		}
		break;
	case AND:
		sprintf(buf, "pand");
		break;
	case ANDN:
		sprintf(buf, "pandn");
		break;
	case OR:
		sprintf(buf, "por");
		break;
	case XOR:
		sprintf(buf, "pxor");
		break;

	case EQ:
		if (tup[s].type.attr & TYP_FLOAT) {
			if (optcpu & CPU_3DNow) {
				sprintf(buf, "pfcmpeq");
			} else {
				bug("float CMPEQ available only for 3DNow!");
			}
		} else {
			snprintf(buf, 256, "pcmpeq%s", show_suf(s));
		}
		break;

	case EQ_C: /* Compare for EQ, with C-like result */
		bug("EQ_C instruction not available for this target");
		break;

	case GT:
		if (tup[s].type.attr & TYP_FLOAT) {
			if (optcpu & CPU_3DNow) {
				sprintf(buf, "pfcmpgt");
			} else {
				bug("float CMPGT available only for 3DNow!");
			}
		} else {
			snprintf(buf, 256, "pcmpgt%s", show_suf(s));
		}
		break;

	case GT_C: /* Compare for GT, with C-like result */
		bug("GT_C instruction not available for this target");
		break;

	case GE:
		if (tup[s].type.attr & TYP_FLOAT) {
			if (optcpu & CPU_3DNow) {
				sprintf(buf, "pfcmpge");
			} else {
				bug("float CMPGE available only for 3DNow!");
			}
		} else {
			bug("GE instruction not available for this target");
		}
		break;

	case SHL:
		snprintf(buf, 256, "psll%s", show_suf(s));
		break;

	case SHR:
		snprintf(buf, 256, "psr%c%s",
			(tup[s].type.attr & TYP_UNSIGN)?
			    'l' :
			    (tup[s].type.attr == TYP_NULL)? 'l' : 'a',
			show_suf(s));
		break;

	case PACK:
		if (tup[s].type.attr & TYP_SAT) {
			snprintf(buf,
				256,
				((tup[s].type.attr & TYP_UNSIGN) ?
				  "packus%s%s" :
				  "packss%s%s"),
				show_suf2(
				  bitsperfield(tup[s].type.bits) * 2),
				show_suf(s));
		} else {
			bug("Generated PACK is incorrect for modular "
			    "types");
		}
		break;

	case UNPACKL:
		snprintf(buf,
			256,
			"punpckl%s%s",
			show_suf2(bitsperfield(tup[s].type.bits) / 2),
			show_suf(s));
		break;
	case UNPACKH:
		snprintf(buf,
			256,
			"punpckh%s%s",
			show_suf2(bitsperfield(tup[s].type.bits) / 2),
			show_suf(s));
		break;

	case MIXEVEN:
		bug("MIXEVEN instruction not available for this target");
		break;

	case MIXODD:
		bug("MIXODD instruction not available for this target");
		break;

	case PERM:
		if (optcpu & CPU_athlon) {
			if (bitsperfield(tup[s].type.bits)==16) {
				snprintf(buf, 256, "pshufw");
			} else {
				bug("Athlon PSHUFW for 16 bits only");
			}
		} else {
			bug("PERM instruction not available for this target");
		}
		break;

	case LEA:
		snprintf(buf, 256, "paddd");
		break;

	case LOADR:
		snprintf(buf, 256, "movq");
		break;

	default:
		{
			char buf[64];
			snprintf(buf,
				64,
				"mmxop op=%s (this cannot happen?)",
				opname(tup[s].op));
			bug(buf);
		}
		return;
	}


	/* Output the arguments */
	#ifdef DEBUG_m2rmode
		fprintf(Cout, "/* m2rmode is %d */", tup[s].m2rmode);
	#endif
	if (tup[s].m2rmode != -1) {
		if (tup[ tup[s].m2rmode ].op == LOAD) {
			if (r0 < REGSAVAIL) {
				fprintf(Cout, "%s_m2r(%s, mm%d);",
					buf,
					symname(tup[ tup[s].m2rmode ].symbol,
						tup[ tup[s].m2rmode ].offset),
					r0);
			} else {
				/* Destination spilled */
				/* Save mm0 */
				fprintf(Cout,
					"movq_r2m(mm0, %s);\n",
					spoolname(REGSAVAIL-1));
				/* Move destination into mm0 */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, mm0);\n",
					spoolname(r0));

				/* Apply op to source and mm0 */
				Ctab();
				fprintf(Cout, "%s_m2r(%s, mm0);",
					buf,
					symname(tup[ tup[s].m2rmode ].symbol,
						tup[ tup[s].m2rmode ].offset));

				/* Move mm0 into destination */
				Ctab();
				fprintf(Cout,
					"movq_r2m(mm0, %s);\n",
					spoolname(r0));
				/* Restore mm0 */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, mm0);",
					spoolname(REGSAVAIL-1));
			}
			#ifdef DEBUG_CODEGEN
				fprintf(Cout," /* %d */\n",__LINE__);
			#endif
		} else {
			/* an immediate from the constant pool */
			switch (tup[s].op) {
			case SHL:
			case SHR:
			case PERM:
				/* do not need constant pool for 8-bit
				   immediates, but this only works for shifts
				   and adds a byte to the opcode, so it is not
				   as clear a win as one might hope...  anyway,
				   use it if we can
				*/
				if (r0 < REGSAVAIL) {
					fprintf(Cout, "%s_i2r(%lld, mm%d);",
					buf,
					(tup[tup[s].m2rmode].immed.q[0] &
					 0xffLL),
					r0);
					#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
					#endif
				} else {
					/* Destination spilled */
					/* Save mm0 */
					fprintf(Cout,
						"movq_r2m(mm0, %s);\n",
						spoolname(REGSAVAIL-1));
					/* Move destination into mm0 */
					Ctab();
					fprintf(Cout,
						"movq_m2r(%s, mm0);\n",
						spoolname(r0));

					/* Apply op to source and mm0 */
					Ctab();
					fprintf(Cout, "%s_i2r(%lld, mm0);\n",
					    buf,
					    (tup[tup[s].m2rmode].immed.q[0] &
					     0xffLL));

					/* Move mm0 into destination */
					Ctab();
					fprintf(Cout,
						"movq_r2m(mm0, %s);\n",
						spoolname(r0));
					/* Restore mm0 */
					Ctab();
					fprintf(Cout,
						"movq_m2r(%s, mm0);",
						spoolname(REGSAVAIL-1));
					#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
					#endif
				}
				break;
			default:
				/* load frag-sized immediates via the constant
				   pool
				*/
				cpoolenter(tup[ tup[s].m2rmode ].immed);

				if (r0 < REGSAVAIL) {
					/* Didn't spill */
				    	fprintf(Cout, "%s_m2r(%s, mm%d);",
					    buf,
					    cpoolname(
						tup[ tup[s].m2rmode ].immed),
					    r0);
					#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
					#endif
				} else {
					/* Destination spilled */
					/* Save mm0 */
					fprintf(Cout,
						"movq_r2m(mm0, %s);\n",
						spoolname(REGSAVAIL-1));
					/* Move destination into mm0 */
					Ctab();
					fprintf(Cout,
						"movq_m2r(%s, mm0);\n",
						spoolname(r0));
					/* Apply op to source and mm0 */
					Ctab();
					fprintf(Cout, "%s_m2r(%s, mm0);\n",
					    buf,
					    cpoolname(
						tup[ tup[s].m2rmode ].immed));
					/* Move mm0 into destination */
					Ctab();
					fprintf(Cout,
						"movq_r2m(mm0, %s);\n",
						spoolname(r0));
					/* Restore mm0 */
					Ctab();
					fprintf(Cout,
						"movq_m2r(%s, mm0);",
						spoolname(REGSAVAIL-1));
					#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
					#endif
				}
			}
		}
	} else {
		if (tup[s].op == LEA) {
			/* Generate the address */
			if (r0 < REGSAVAIL) {
				fprintf(Cout, "%s_m2r((p64_t *) &%s, mm%d);",
					buf, tup[s].symbol->text, r0);
			} else {
				/* Destination spilled */
				/* Save mm0 */
				fprintf(Cout,
					"movq_r2m(mm0, %s);\n",
					spoolname(REGSAVAIL-1));
				/* Move destination into mm0 */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, mm0);\n",
					spoolname(r0));

				/* Apply op to source and mm0 */
				Ctab();
				fprintf(Cout, "%s_m2r((p64_t *) &%s, mm0);\n",
					buf,
					tup[s].symbol->text);

				/* Move mm0 into destination */
				Ctab();
				fprintf(Cout,
					"movq_r2m(mm0, %s);\n",
					spoolname(r0));
				/* Restore mm0 */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, mm0);",
					spoolname(REGSAVAIL-1));
			}
			#ifdef DEBUG_CODEGEN
				fprintf(Cout,"/* %d */\n",__LINE__);
			#endif
		} else if (tup[s].op == LOADR) {
			/* Generate the load using address in register */
			if (r0 < REGSAVAIL) {
				fprintf(Cout, "%s_x2r(mm%d, mm%d);",
					buf, r0, r0);
			} else {
				/* Spilled */
				/* Save mm0 */
				fprintf(Cout,
					"movq_r2m(mm0, %s);\n",
					spoolname(REGSAVAIL-1));
				/* Move destination into mm0 */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, mm0);\n",
					spoolname(r0));
				/* Apply op to source and mm0 (as dest) */
				Ctab();
				fprintf(Cout,
					"%s_x2r(mm0, mm0);",
					buf);
				/* Move mm0 into destination */
				Ctab();
				fprintf(Cout,
					"movq_r2m(mm0, %s);\n",
					spoolname(r0));
				/* Restore mm0 */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, mm0);",
					spoolname(REGSAVAIL-1));
			}
			#ifdef DEBUG_CODEGEN
				fprintf(Cout,"/* %d */\n",__LINE__);
			#endif
		} else {
			if ( (r1 < REGSAVAIL) && (r0 < REGSAVAIL) ) {
				/* Neither spilled */
				fprintf(Cout, "%s_r2r(mm%d, mm%d);",
					buf, r1, r0);
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,"/* %d */\n",__LINE__);
				#endif
			} else if ( (r1 >= REGSAVAIL) && (r0 < REGSAVAIL) ) {
				/* Source spilled */
				fprintf(Cout, "%s_m2r(%s, mm%d);",
					buf,
					spoolname(r1),
					r0);
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,"/* %d */\n",__LINE__);
				#endif
			} else if ( (r1 < REGSAVAIL) && (r0 >= REGSAVAIL) ) {
				/* Destination spilled */
				int savereg = (r1==0)? 1:0;

				/* Save a register */
				fprintf(Cout,
					"movq_r2m(mm%d, %s);\n",
					savereg,
					spoolname(REGSAVAIL-1));
				/* Move destination into savereg */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, mm%d);\n",
					spoolname(r0),
					savereg);
				/* Apply op to source and savereg (as dest) */
				Ctab();
				fprintf(Cout,
					"%s_r2r(mm%d, mm%d);\n",
					buf,
					r1,
					savereg);
				/* Move savereg into destination */
				Ctab();
				fprintf(Cout,
					"movq_r2m(mm%d, %s);\n",
					savereg,
					spoolname(r0));
				/* Restore savereg */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, mm%d);",
					spoolname(REGSAVAIL-1),
					savereg);
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,"/* %d */\n",__LINE__);
				#endif
			} else if ( (r1 >= REGSAVAIL) && (r0 >= REGSAVAIL) ) {
				/* Source and destination spilled */
				/* Save mm0 */
				fprintf(Cout,
					"movq_r2m(mm0, %s);\n",
					spoolname(REGSAVAIL-1));
				/* Move destination into mm0 */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, mm0);\n",
					spoolname(r0));
				/* Apply op to source and mm0 (as dest) */
				Ctab();
				fprintf(Cout,
					"%s_m2r(%s, mm0);\n",
					buf,
					spoolname(r1));
				/* Move mm0 into destination */
				Ctab();
				fprintf(Cout,
					"movq_r2m(mm0, %s);\n",
					spoolname(r0));
				/* Restore mm0 */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, mm0);",
					spoolname(REGSAVAIL-1));
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,"/* %d */\n",__LINE__);
				#endif
			}
		}
	}
	fprintf(Cout, "\n");
}

static void
maxop(register int s)
{
	register int r0 = -1;
	register int r1 = -1;
	char buf[256];		/* This is much larger than it needs to be */

	/* Get first arg, with possible rename */
	if (tup[s].arg[0] != -1) {
		r0 = tup[ tup[s].arg[0] ].reg;

		if (tup[s].antidep != -1) {
			/* Arg was renamed; output a move */
			r0 = tup[s].reg;
			Ctab();

			if ( (r0 < REGSAVAIL) &&
			     (tup[s].antidep < REGSAVAIL) ) {
				/* Neither spilled */
				fprintf(Cout,
					"movq_r2r(r%d, r%d);\n",
					r0,
					tup[s].antidep);
			#ifdef DEBUG_CODEGEN
				fprintf(Cout,"/* %d */\n",__LINE__);
			#endif
			} else if ( (r0 >= REGSAVAIL) &&
			     (tup[s].antidep < REGSAVAIL) ) {
				/* Source spilled */
				fprintf(Cout,
					"movq_m2r(%s, r%d);\n",
					spoolname(r0),
					tup[s].antidep);
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,
						"/* %d */\n",
						__LINE__);
				#endif
			} else if ( (r0 < REGSAVAIL) &&
			     (tup[s].antidep >= REGSAVAIL) ) {
				/* Destination spilled */
				fprintf(Cout,
					"movq_r2m(r%d, %s);\n",
					r0,
					spoolname(tup[s].antidep));
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,
						"/* %d */\n",
						__LINE__);
				#endif
			} else if ( (r0 >= REGSAVAIL) &&
			     (tup[s].antidep >= REGSAVAIL) ) {
				/* Source and destination spilled */
				/* Save r0 */
				fprintf(Cout,
					"movq_r2m(r0, %s);\n",
					spoolname(REGSAVAIL-1));
				/* Move source into r0 */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, r0);\n",
					spoolname(r0));
				/* Move source into destination */
				Ctab();
				fprintf(Cout,
					"movq_r2m(r0, %s);\n",
					spoolname(tup[s].antidep));
				/* Restore r0 */
				Ctab();
				fprintf(Cout,
					"movq_m2r(%s, r0);\n",
					spoolname(REGSAVAIL-1));
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,
						"/* %d */\n",
						__LINE__);
				#endif
			}
			tup[ tup[s].arg[0] ].reg = tup[s].antidep;
		}
	}

	/* Get second arg */
	if (tup[s].arg[1] != -1) {
		r1 = tup[ tup[s].arg[1] ].reg;
	}

	/* Where does result go? */
	tup[s].reg = r0;

	Ctab();
	switch (tup[s].op) {
	case ADD:
		if (tup[s].type.attr & TYP_FLOAT) {
			bug("float ADD available only for 3DNow!");
		} else {
			if (strcmp(show_suf(s), "q")) {
				snprintf(buf, 256, "hadd%s%s",
				 (tup[s].type.attr & TYP_SAT)?
				  ((tup[s].type.attr & TYP_UNSIGN)?
				   "us": "ss"):
				  (""),
				 show_suf(s));
			} else {
				snprintf(buf, 256, "add%s",
				 (tup[s].type.attr & TYP_SAT)?
				  ((tup[s].type.attr & TYP_UNSIGN)?
				   "us": "ss"):
				  (""));
			}
		}
		break;
	case SUB:
		if (strcmp(show_suf(s), "q")) {
			snprintf(buf, 256, "hsub%s%s",
			 (tup[s].type.attr & TYP_SAT)?
			  ((tup[s].type.attr & TYP_UNSIGN)?
			   "us": "ss"):
			  (""),
			 show_suf(s));
		} else {
			snprintf(buf, 256, "sub%s",
			 (tup[s].type.attr & TYP_SAT)?
			  ((tup[s].type.attr & TYP_UNSIGN)?
			   "us": "ss"):
			  (""));
		}
		break;
	case MUL:
		bug("MUL instruction not available for this target");
		break;
	case MULH:
		bug("MULH instruction not available for this target");
		break;
	case DIV:
		snprintf(buf,
			256,
			"pdiv%s%s",
			((tup[s].type.attr & TYP_UNSIGN) ? "u" : ""),
			show_suf(s));
		break;
	case MOD:
		snprintf(buf,
			256,
			"pmod%s%s",
			((tup[s].type.attr & TYP_UNSIGN) ? "u" : ""),
			show_suf(s));
		break;
	case AVG:
		snprintf(buf, 256, "havg%s%s",
			 (tup[s].type.attr & TYP_SAT)?
			  ((tup[s].type.attr & TYP_UNSIGN)?
			   "us": "ss"):
			  (""),
			 show_suf(s));
		break;

	case MAX:
		bug("MAX instruction not available on this target");
		break;
	case MIN:
		bug("MIN instruction not available for this target");
		break;
	case AND:
		sprintf(buf, "and");
		break;
	case ANDN:
		sprintf(buf, "andcm");
		break;
	case OR:
		sprintf(buf, "or");
		break;
	case XOR:
		sprintf(buf, "xor");
		break;

	case EQ:
		snprintf(buf,
			256,
			"cmpeq%s%s",
			(tup[s].type.attr & TYP_UNSIGN)? "u":"",
			show_suf(s));
		break;

	case GT:
		snprintf(buf,
			256,
			"cmpgt%s%s",
			(tup[s].type.attr & TYP_UNSIGN)? "u":"",
			show_suf(s));
		break;

	case GT_C: /* Compare for GT, with C-like result */
		bug("GT_C instruction not available for this target");
		break;

	case SHL:
		sprintf(buf, "hshl");
		break;

	case SHR:
		sprintf(buf, "hshr%s",
			(tup[s].type.attr & TYP_UNSIGN)? "u" : "");
		break;

	case PACK:
		bug("PACK instruction not available for this target");
		break;

	case UNPACKL:
		bug("PUNPCKL instruction not available for this target");
		break;

	case UNPACKH:
		bug("PUNPCKH instruction not available for this target");
		break;

	case MIXEVEN:
		snprintf(buf, 256, "mixr%s", show_suf(s));
		break;

	case MIXODD:
		snprintf(buf, 256, "mixl%s", show_suf(s));
		break;

	case PERM:
		snprintf(buf, 256, "perm%s", show_suf(s));
		break;

	default:
		{
			char buf[64];
			snprintf(buf,
				64,
				"maxop op=%s (this cannot happen?)",
				opname(tup[s].op));
			bug(buf);
		}
		return;
	}


	/* Output the arguments */
	if (tup[s].m2rmode != -1) {
		if (tup[ tup[s].m2rmode ].op == LOAD) {
				fprintf(Cout, "%s_m2r(%s, r%d);",
					buf,
					symname(tup[ tup[s].m2rmode ].symbol,
						tup[ tup[s].m2rmode ].offset),
					r0);
				#ifdef DEBUG_CODEGEN
					fprintf(Cout,"/* %d */\n",__LINE__);
				#endif
		} else {
			/* an immediate from the constant pool */
			switch (tup[s].op) {
			case SHL:
			case SHR:
			case PERM:
				/* do not need constant pool for 8-bit
				   immediates, but this only works for shifts
				   and adds a byte to the opcode, so it is not
				   as clear a win as one might hope...  anyway,
				   use it if we can
				*/
				if (r0 < REGSAVAIL) {
					fprintf(Cout, "%s_i2r(%lld, r%d);",
					buf,
					(tup[tup[s].m2rmode].immed.q[0] &
					 0xffLL),
					r0);
					#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
					#endif
				} else {
					/* Destination spilled */
					/* Save r0 */
					fprintf(Cout,
						"movq_r2m(r0, %s);\n",
						spoolname(REGSAVAIL-1));
					/* Move destination into r0 */
					Ctab();
					fprintf(Cout,
						"movq_m2r(%s, r0);\n",
						spoolname(r0));

					/* Apply op to source and r0 */
					Ctab();
					fprintf(Cout, "%s_i2r(%lld, r0);\n",
					    buf,
					    (tup[tup[s].m2rmode].immed.q[0] &
					     0xffLL));

					/* Move r0 into destination */
					Ctab();
					fprintf(Cout,
						"movq_r2m(r0, %s);\n",
						spoolname(r0));
					/* Restore r0 */
					Ctab();
					fprintf(Cout,
						"movq_m2r(%s, r0);",
						spoolname(REGSAVAIL-1));
					#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
					#endif
				}
				break;
			default:
				/* load frag-sized immediates via the constant
				   pool
				*/
				cpoolenter(tup[ tup[s].m2rmode ].immed);

				if (r0 < REGSAVAIL) {
					/* Didn't spill */
				    	fprintf(Cout, "%s_m2r(%s, r%d);",
					    buf,
					    cpoolname(
						tup[ tup[s].m2rmode ].immed),
					    r0);
					#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
					#endif
				} else {
					/* Destination spilled */
					/* Save r0 */
					fprintf(Cout,
						"movq_r2m(r0, %s);\n",
						spoolname(REGSAVAIL-1));
					/* Move destination into r0 */
					Ctab();
					fprintf(Cout,
						"movq_m2r(%s, r0);\n",
						spoolname(r0));
					/* Apply op to source and r0 */
					Ctab();
					fprintf(Cout, "%s_m2r(%s, r0);\n",
					    buf,
					    cpoolname(
						tup[ tup[s].m2rmode ].immed));
					/* Move r0 into destination */
					Ctab();
					fprintf(Cout,
						"movq_r2m(r0, %s);\n",
						spoolname(r0));
					/* Restore r0 */
					Ctab();
					fprintf(Cout,
						"movq_m2r(%s, r0);",
						spoolname(REGSAVAIL-1));
					#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
					#endif
				}
			}
		}
	} else {
		if ( (r1 < REGSAVAIL) && (r0 < REGSAVAIL) ) {
			/* Neither spilled */
			fprintf(Cout, "%s_r2r(r%d, r%d);",
				buf, r1, r0);
			#ifdef DEBUG_CODEGEN
				fprintf(Cout,"/* %d */\n",__LINE__);
			#endif
		} else if ( (r1 >= REGSAVAIL) && (r0 < REGSAVAIL) ) {
			/* Source spilled */
			fprintf(Cout, "%s_m2r(%s, r%d);",
				buf,
				spoolname(r1),
				r0);
			#ifdef DEBUG_CODEGEN
				fprintf(Cout,"/* %d */\n",__LINE__);
			#endif
		} else if ( (r1 < REGSAVAIL) && (r0 >= REGSAVAIL) ) {
			/* Destination spilled */
			int savereg = (r1==0)? 1:0;

			/* Save a register */
			fprintf(Cout,
				"movq_r2m(r%d, %s);\n",
				savereg,
				spoolname(REGSAVAIL-1));
			/* Move destination into savereg */
			Ctab();
			fprintf(Cout,
				"movq_m2r(%s, r%d);\n",
				spoolname(r0),
				savereg);
			/* Apply op to source and savereg (as dest) */
			Ctab();
			fprintf(Cout,
				"%s_r2r(r%d, r%d);\n",
				buf,
				r1,
				savereg);
			/* Move savereg into destination */
			Ctab();
			fprintf(Cout,
				"movq_r2m(r%d, %s);\n",
				savereg,
				spoolname(r0));
			/* Restore savereg */
			Ctab();
			fprintf(Cout,
				"movq_m2r(%s, r%d);",
				spoolname(REGSAVAIL-1),
				savereg);
			#ifdef DEBUG_CODEGEN
				fprintf(Cout,"/* %d */\n",__LINE__);
			#endif
		} else if ( (r1 >= REGSAVAIL) && (r0 >= REGSAVAIL) ) {
			/* Source and destination spilled */
			/* Save r0 */
			fprintf(Cout,
				"movq_r2m(r0, %s);\n",
				spoolname(REGSAVAIL-1));
			/* Move destination into r0 */
			Ctab();
			fprintf(Cout,
				"movq_m2r(%s, r0);\n",
				spoolname(r0));
			/* Apply op to source and r0 (as dest) */
			Ctab();
			fprintf(Cout,
				"%s_m2r(%s, r0);\n",
				buf,
				spoolname(r1));
			/* Move r0 into destination */
			Ctab();
			fprintf(Cout,
				"movq_r2m(r0, %s);\n",
				spoolname(r0));
			/* Restore r0 */
			Ctab();
			fprintf(Cout,
				"movq_m2r(%s, r0);",
				spoolname(REGSAVAIL-1));
			#ifdef DEBUG_CODEGEN
				fprintf(Cout,"/* %d */\n",__LINE__);
			#endif
		}
	}
	fprintf(Cout, "\n");
}

static int
possible_conflict(register int i,
register int j)
{
	/* Returns 1 if tuples i and j may be memory accesses of
	   the same fragment of the same vector; 0 otherwise.
	*/
	sym *s;
	int o;

	switch(tup[i].op) {
	case LOAD:
		s = tup[i].symbol;
		o = tup[i].offset;

		switch(tup[j].op) {
		case LOAD:
			return 0;
		case LOADR:
			if (s == tup[tup[j].arg[0]].symbol) {
				return 1;
			} else {
				return 0;
			}
		case STORE:
			if ((s == tup[j].symbol) && (o == tup[j].offset)) {
				return 1;
			} else {
				return 0;
			}
		case STORER:
			if (s == tup[tup[j].arg[0]].symbol) {
				return 1;
			} else {
				return 0;
			}
		default:
			return 0;
		}
		break;

	case LOADR:
		s = tup[tup[i].arg[0]].symbol;

		switch(tup[j].op) {
		case LOAD:
			if (s == tup[j].symbol) {
				return 1;
			} else {
				return 0;
			}
		case LOADR:
			if (s == tup[tup[j].arg[0]].symbol) {
				return 1;
			} else {
				return 0;
			}
		case STORE:
			if (s == tup[j].symbol) {
				return 1;
			} else {
				return 0;
			}
		case STORER:
			if (s == tup[tup[j].arg[0]].symbol) {
				return 1;
			} else {
				return 0;
			}
		default:
			return 0;
		}
		break;

	case STORE:
		s = tup[i].symbol;

		switch(tup[j].op) {
		case LOAD:
			if ((s == tup[j].symbol) && (o == tup[j].offset)) {
				return 1;
			} else {
				return 0;
			}
		case LOADR:
			if (s == tup[tup[j].arg[0]].symbol) {
				return 1;
			} else {
				return 0;
			}
		case STORE:
			if (s == tup[j].symbol) {
				return 1;
			} else {
				return 0;
			}
		case STORER:
			if (s == tup[tup[j].arg[0]].symbol) {
				return 1;
			} else {
				return 0;
			}
		default:
			return 0;
		}
		break;

	case STORER:
		s = tup[tup[i].arg[0]].symbol;
		switch(tup[j].op) {
		case LOAD:
			if (s == tup[j].symbol) {
				return 1;
			} else {
				return 0;
			}
		case LOADR:
			if (s == tup[tup[j].arg[0]].symbol) {
				return 1;
			} else {
				return 0;
			}
		case STORE:
			if (s == tup[j].symbol) {
				return 1;
			} else {
				return 0;
			}
		case STORER:
			if (s == tup[tup[j].arg[0]].symbol) {
				return 1;
			} else {
				return 0;
			}
		default:
			return 0;
		}
		break;

	default:
		return 0;
	}
}

static int
storeord1(register int root,
register int pos)
{
	register int i;

	if (root == -1) return(pos);

	/* Also done if scheduled earlier */
	for (i=0; i<pos; ++i) {
		if (sbest[i] == root) return(pos);
	}

	/* Pull arguments ahead of root */
	pos = storeord1(tup[root].arg[0], pos);
	pos = storeord1(tup[root].arg[1], pos);

	/* Now put root here */
	sbest[pos] = root;

	/* And we've added another */
	return(pos + 1);
}

static void
storeord()
{
	/* order tuples by back-tracing stores */
	/* Does NOT allocate registers */

	register int i, j, k;
	register int stores;

	#ifdef DEBUG_SCHED
		fprintf(stderr, "storeord() schedsp=%d\n", schedsp);
	#endif

	/* move stores to front while maintaining order */
	stores = 0;
	for (i=0; i<schedsp; ++i) {
		if ((tup[ sched[i] ].op == STORE) ||
		    (tup[ sched[i] ].op == STORER)) {
			k = sched[i];
			sched[i] = sched[stores];
			sched[stores] = k;
			++stores;
		}
	}

	/* sort stores into increasing "before" value, without changing the
	   order of stores that may be to the same fragment. */
	/* This needs to be fixed a little: if the swap is disallowed because
	   of a possible conflict between i and some k, but there is no
	   conflict between i and j or between any k and j, then j should be
	   floated up before i if j has a smaller "before" value.
	*/
	for (i=0; i<stores; ++i) {
	    for (j=i+1; j<stores; ++j) {
		if (tup[ sched[i] ].before > tup[ sched[j] ].before) {
			int disallowed = possible_conflict(sched[i], sched[j]);
			int swap_disallowed = 0;

			for ( k=i+1; ((!disallowed) && (k<j)); ++k)
			{
			    disallowed = possible_conflict(sched[j], sched[k]);
			    swap_disallowed =
					swap_disallowed ||
					possible_conflict(sched[i], sched[k]) ||
					disallowed;
			}

			if (!disallowed) {
			    if (!swap_disallowed) {
				/* Just swap i and j */
				k = sched[i];
				sched[i] = sched[j];
				sched[j] = k;
			    } else {
				/* Promote j to i without changing i through
				   j-1's relative position.
					(e.g. 4,5,6,7 -> 7,4,5,6)
				*/
				int tmp = sched[j];
				for (k=j; k>i; --k) {
					sched[k] = sched[k-1];
				}
				sched[i] = tmp;
			    }
			}
		}
	    }
	}

	/* schedule tree under each store */
	j = 0;
	for (i=0; i<stores; ++i) {
		j = storeord1(sched[i], j);
	}

	if (j != schedsp) {
		bug("basic block prescheduler failed");
	}

	for (i=0; i<schedsp; ++i) {
		sched[i] = sbest[i];
	}
}

void dump_tuples(FILE *strm, char *msg)
{
	int i;
	tuple *t;

	fprintf (strm, "Dumping tuple list %s:\n", msg);
	for (i=0; i<tupsp; ++i) {
		t = &(tup[i]);
		/* This is split into multiple calls because the version of
		   gcc on my laptop computer doesn't handle 64-bit function
		   arguments properly and screws up the values after t->immed.
		   This has probably been fixed in later versions of gcc.
		*/
		fprintf(strm,
			" %4d:%s args(%d,%d) \"%s\" [%d] (%d)\n",
			i, opname(t->op),
			t->arg[0], t->arg[1],
			((t->symbol)? t->symbol->text : "(null)"),
			tup[t->offset].immed,
			t->immed);
		fprintf(strm,
			"      ad=%d rfs=%d,%d dpth=%d bfr/aft=%d,%d",
			t->antidep,
			t->refs, t->trefs,
			t->depth,
			t->before, t->after);
		fprintf(strm,
			" reg=%d,%d m2r=%d spld=%d splrg=%d\n",
			t->reg, t->oreg,
			t->m2rmode,
			t->spilled, t->spillreg);
	}
}

void dump_sched(FILE *strm, char *msg)
{
	int i;
	tuple *t;

	fprintf (strm, "Dumping schedule %s:\n", msg);
	for (i=0; i<schedsp; ++i) {
		t = &(tup[sched[i]]);
		fprintf(strm,
			" %4d[%d]:%s args(%d,%d) \"%s\" [%d] (%d)\n",
			i, sched[i], opname(t->op),
			t->arg[0], t->arg[1],
			((t->symbol)? t->symbol->text : "(null)"),
			tup[t->offset].immed,
			t->immed);
		fprintf(strm,
			"     ad=%d rfs/t=%d,%d dpth=%d bfr/aft=%d,%d",
			t->antidep,
			t->refs, t->trefs,
			t->depth,
			t->before, t->after);
		fprintf(strm,
			" reg/o=%d,%d m2r=%d spld=%d splrg=%d\n",
			t->reg, t->oreg,
			t->m2rmode,
			t->spilled, t->spillreg);
	}
}

#ifdef NEWSTUFF
static int tos;
void sched_node(int i)
{
	#ifdef DEBUG_SEARCH
		fprintf(stderr, "Start sched_node(%d)\n", i);
		fprintf(stderr, "tup[%d].refs=%d, tup[%d].reg=%d\n",
				i, tup[i].refs, i, tup[i].reg);
		fflush(stderr);
	#endif

	if ((tup[i].refs < 1) || (tup[i].reg != 0)) return;

	switch (tup[i].op) {

	/* Nullary (What, if anything, is latin for zero?) ops */
	case NUM:
	case LOAD:	tup[i].reg = tos++;
			/* sched[schedsp++] = i; */
			break;

	/* Unary ops not requiring register for result */
	case STORE:	sched_node(tup[i].arg[0]);
			/* sched[schedsp++] = i; */
			break;

	/* Unary ops requiring register for result */
	case I2F:
	case F2I:
	case LNOT:
	case NEG:
	case NOT:
	case RCP:
			sched_node(tup[i].arg[0]);
			tup[i].reg = (tup[tup[i].arg[0]].refs == 1)?
					tup[tup[i].arg[0]].reg : tos++;
			/* sched[schedsp++] = i; */
			break;

	/* Binary ops */
	case ADD:
	case AND:
	case ANDN:
	case AVG:
	case DIV:
	case EQ:
	case EQ_C:
	case GE:
	case GT:
	case GT_C:
	case MOD:
	case MUL:
	case MULH:
	case OR:
	case SHL:
	case SHR:
	case SUB:
	case XOR:
	case PACK:
	case UNPACKL:
	case UNPACKH:
	case MIXEVEN:
	case MIXODD:
	case PERM:
	case MAX:
	case MIN:
	case RCP1:
	case RCP2:
			if (tup[tup[i].arg[0]].arg[1] == -1) {
				sched_node(tup[i].arg[1]);
				sched_node(tup[i].arg[0]);
			} else {
				sched_node(tup[i].arg[0]);
				sched_node(tup[i].arg[1]);
			}
			tup[i].reg = (tup[tup[i].arg[0]].refs == 1)?
					tup[tup[i].arg[0]].reg : tos++;
			/* sched[schedsp++] = i; */
			break;
	default:
		{
			char buf[256];
			sprintf(buf,
				"sched_node(%s) defaulted",
				opname(tup[i].op));
			error(buf);
		}
	}
	#ifdef DEBUG_SEARCH
		fprintf(stderr, "End sched_node(%d)\n", i);
		fflush(stderr);
	#endif
}

static int
sched_simple(void)
{
	int i;

	#ifdef DEBUG_SEARCH
		fprintf(stderr, "Start sched_simple()\n");
		fprintf(stderr, "schedsp=%d, tupsp=%d\n", schedsp, tupsp);
		fflush(stderr);
	#endif

	/* Set the number of allowed spills */
	numspills = MAXSPILLS;

	#ifdef DEBUG_SEARCH
		dump_tuples(stderr, "before storeord(simple)");
		fprintf(stderr, "schedsp=%d, tupsp=%d\n", schedsp, tupsp);
		fflush(stderr);
	#endif

	/* Mark tuples as unspilled, and Make schedule index */
	schedsp = 0;
	for (i=0; i<tupsp; ++i) {
		if (tup[i].refs > 0) {
			tup[i].spilled = -1;
			sched[schedsp++] = i;
		}
	}

	storeord();

	#ifdef DEBUG_SEARCH
		dump_sched(stderr, "after storeord(simple)");
		fprintf(stderr, "schedsp=%d, tupsp=%d\n", schedsp, tupsp);
		fflush(stderr);
	#endif

	/* Search for a good schedule */
	sbestval = NOTFOUND;
	schedval = 0;

	tos = 0;
	for (i=0; i<tupsp; ++i) sched_node(i);

	for (i=0; i<schedsp; ++i) schedval += sched_cost(i);
	sched_eval();

	#ifdef DEBUG_SEARCH
		dump_sched(stderr, "after sched_simple()");
		fprintf(stderr, "sbestval = %d\n", sbestval);
		fflush(stderr);
	#endif

	return sbestval;
}
#endif

static void
refcounts(void)
{
	register int i;

	/* Set reference counts */
	for (i=0; i<tupsp; ++i) {
		if ((tup[i].op == STORE) && (tup[i].refs > 0)) {
			incref(tup[i].arg[0]);
		} else if ((tup[i].op == STORER) && (tup[i].refs > 0)) {
			incref(tup[i].arg[0]);
			incref(tup[i].arg[1]);
		} else {
			tup[i].refs = 0;
		}
	}
}


static int
sched_search(int allowed)
{
	register int i, j, k;

	#ifdef DEBUG_SEARCH
	{
		char buf[256];
		snprintf(buf, 256, "Starting sched_search(%d)", allowed);
		info(0, buf);
	}
	#endif

	/* Set the number of allowed spills */
	numspills = allowed;

	/* Mark tuples as unspilled, and make schedule index */
	schedsp = 0;
	for (i=0; i<tupsp; ++i) {
		if (tup[i].refs > 0) {
			tup[i].spilled = -1;
			sched[schedsp++] = i;
		}
	}

	#ifdef DEBUG_SEARCH
		dump_tuples(stderr, "before storeord()");
	#endif

	/* Create "initial schedule" that will help
	   the search find better schedules quicker...
	   This schedule will not have register allocation.
	*/
	#ifdef DEBUG_SEARCH
		info(0, "creating initial schedule");
	#else
		info(2, "creating initial schedule");
	#endif
	storeord();

	#ifdef DEBUG_SEARCH
		dump_sched(stderr, "after storeord()");
	#endif

	/* Search for a good schedule */
	sbestval = NOTFOUND;
	schedtim = (time(0) + opttime);
	schedval = 0;
	sched_perm(0);

	#ifdef DEBUG_SEARCH
		dump_sched(stderr, "after sched_perm()");
		fprintf(stderr, "sbestval = %d\n", sbestval);
	#endif

	if (optm2r || ((optcpu & CPU_RISC) == 0)) {
		/* force using non-RISC addressing forms */
		sbestval = NOTFOUND;
	}



	if (sbestval == NOTFOUND) {
		/* Try again, this time using m2r single-refs and i2r shifts */
		#ifdef DEBUG_SEARCH
			info(0, "using m2r for single refs and "
				"i2r for shifts to free registers");
		#else
			info(2, "using m2r for single refs and i2r for shifts "
				"and permutes to free registers");
		#endif
		for (i=0; i<tupsp; ++i) {
			if ((tup[i].refs > 0) &&
			    (tup[i].op != STORE) &&
			    (tup[i].op != STORER)) {
				register int j = tup[i].arg[1];

				if (j != -1) {
					if (((tup[i].op == SHL) ||
					     (tup[i].op == SHR) ||
					     (tup[i].op == PERM)) &&
					    (tup[j].op == NUM)) {
						tup[i].m2rmode = j;
						--(tup[j].refs);
						tup[i].arg[1] = -1;
					} else if ((tup[j].refs == 1) &&
						   ((tup[j].op == NUM) ||
						    (tup[j].op == LOAD))) {
						tup[i].m2rmode = j;
						--(tup[j].refs);
						tup[i].arg[1] = -1;
					}

					j = tup[i].arg[0];
					if ((j != -1) &&
					    (tup[i].m2rmode == -1) &&
					    (tup[j].refs == 1) &&
					    (!ordered(tup[i].op)) &&
					    ((tup[j].op == NUM) ||
					     (tup[j].op == LOAD))) {
					     /* swap args and absorb constant */
						tup[i].m2rmode = j;
						tup[j].refs = 0;
						tup[i].arg[0] = tup[i].arg[1];
						tup[i].arg[1] = -1;
					}
				}
			}
		}

		#ifdef DEBUG_SEARCH
			info(0, "out of first loop");
		#endif

		/* Set reference counts */
		refcounts();

		/* Now do before and after fields */
		before_bb(0,0,1);	/* Just inits the static cache */
		for (i=0; i<tupsp; ++i) {
			if (tup[i].refs > 0) {
				tup[i].before = 0;
				tup[i].after = 0;
				for (j=0; j<tupsp; ++j) if (i != j) {
					if (before_bb(j, i, 0)) {
						++(tup[i].before);
					} else if (before_bb(i, j, 0)) {
						++(tup[i].after);
					}
				}
			}
		}

		#ifdef DEBUG_SEARCH
			info(0, "done setting before and after fields");
		#endif


		/* Make schedule index */
		schedsp = 0;
		for (i=0; i<tupsp; ++i) {
			if (tup[i].refs > 0) {
				tup[i].spilled = -1;
				sched[schedsp++] = i;
			}
		}

		#ifdef DEBUG_SEARCH
			info(0, "done making schedule index");
		#endif

		/* Search for a good schedule */
		sbestval = NOTFOUND;
		schedtim = (time(0) + opttime);
		schedval = 0;
		#ifdef DEBUG_SEARCH
			dump_sched(stderr, "before sched_perm()");
			fprintf(stderr, "sbestval = %d\n", sbestval);
			info(0, "calling sched_perm...");
		#endif

		sched_perm(0);
		#ifdef DEBUG_SEARCH
			info(0, "...back from sched_perm");
			dump_sched(stderr, "after sched_perm()");
			fprintf(stderr, "sbestval = %d\n", sbestval);
		#endif
	}



	if (sbestval == NOTFOUND) {
		/* Try again, this time using m2r constants */
		#ifdef DEBUG_SEARCH
			info(0, "using m2r constants to free registers");
		#else
			info(2, "using m2r constants to free registers");
		#endif
		for (i=0; i<tupsp; ++i) {
			if ((tup[i].refs > 0) &&
			    (tup[i].op != STORE) &&
			    (tup[i].op != STORER)) {
				register int j = tup[i].arg[1];

				if (j != -1) {
					if (tup[j].op == NUM) {
						tup[i].m2rmode = j;
						--(tup[j].refs);
						tup[i].arg[1] = -1;
					}

					j = tup[i].arg[0];
					if ((j != -1) &&
					    (tup[i].m2rmode == -1) &&
					    (!ordered(tup[i].op)) &&
					    (tup[ j ].op == NUM)) {
						/* swap args and absorb
						   constant */
						tup[i].m2rmode = j;
						tup[j].refs = 0;
						tup[i].arg[0] = tup[i].arg[1];
						tup[i].arg[1] = -1;
					}
				}
			}
		}

		/* Set reference counts */
		refcounts();

		/* Now do before and after fields */
		before_bb(0,0,1);	/* Just inits the static cache */
		for (i=0; i<tupsp; ++i) {
			if (tup[i].refs > 0) {
				tup[i].before = 0;
				tup[i].after = 0;
				for (j=0; j<tupsp; ++j) if (i != j) {
					if (before_bb(j, i, 0))
						++(tup[i].before);
					else if (before_bb(i, j, 0))
						++(tup[i].after);
				}
			}
		}

		/* Make schedule index */
		schedsp = 0;
		for (i=0; i<tupsp; ++i) {
			if (tup[i].refs > 0) {
				tup[i].spilled = -1;
				sched[schedsp++] = i;
			}
		}

		/* Search for a good schedule */
		sbestval = NOTFOUND;
		schedtim = (time(0) + opttime);
		schedval = 0;
		sched_perm(0);
	}



	if (sbestval == NOTFOUND) {
		/* Try again, this time using m2r loads */
		#ifdef DEBUG_SEARCH
			info(0, "using m2r loads to free registers");
		#else
			info(2, "using m2r loads to free registers");
		#endif
		for (i=0; i<tupsp; ++i) {
			if ((tup[i].refs > 0) &&
			    (tup[i].op != STORE) &&
			    (tup[i].op != STORER)) {
				register int j = tup[i].arg[1];

				if (j != -1) {
					if (tup[j].op == LOAD) {
						tup[i].m2rmode = j;
						--(tup[j].refs);
						tup[i].arg[1] = -1;
					}

					j = tup[i].arg[0];
					if ((j != -1) &&
					    (tup[i].m2rmode == -1) &&
					    (!ordered(tup[i].op)) &&
					    (tup[ j ].op == LOAD)) {
					     /* swap args and absorb constant */
						tup[i].m2rmode = j;
						tup[j].refs = 0;
						tup[i].arg[0] = tup[i].arg[1];
						tup[i].arg[1] = -1;
					}
				}
			}
		}

		/* Set reference counts */
		refcounts();

		/* Now do before and after fields */
		before_bb(0,0,1);	/* Just inits the static cache */
		for (i=0; i<tupsp; ++i) {
			if (tup[i].refs > 0) {
				tup[i].before = 0;
				tup[i].after = 0;
				for (j=0; j<tupsp; ++j) if (i != j) {
					if (before_bb(j, i, 0))
						++(tup[i].before);
					else if (before_bb(i, j, 0))
						++(tup[i].after);
				}
			}
		}

		/* Make schedule index */
		schedsp = 0;
		for (i=0; i<tupsp; ++i) {
			if (tup[i].refs > 0) {
				tup[i].spilled = -1;
				sched[schedsp++] = i;
			}
		}

		/* Search for a good schedule */
		sbestval = NOTFOUND;
		schedtim = (time(0) + opttime);
		schedval = 0;
		sched_perm(0);
	}

	return sbestval;
}

void
output_schedule(void)
{
	register int i;

		for (i=0; i<schedsp; ++i) {
			register int s = sched[i];

			if (argregs(i)) {
				bug("scheduled register allocation was not "
				    "valid (this cannot happen?)");
			}

			/* write final tuple order to file */
			switch (tup[s].op) {
			case LOAD:
				Ctab();
				if (optcpu & CPU_MMX) {
					if (tup[s].oreg < REGSAVAIL) {
						/* Didn't spill */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						fprintf(Cout,
						"movq_m2r(%s, mm%d);\n",
						symname(tup[s].symbol,
							tup[s].offset),
						(tup[s].reg = tup[s].oreg));
					} else {
						/* Destination spilled */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						/* Save mm0 */
						fprintf(Cout,
							"movq_r2m(mm0, %s);\n",
							spoolname(REGSAVAIL-1));
						/* Move source into r0 */
						Ctab();
						fprintf(Cout,
						"movq_m2r(%s, mm0);\n",
						symname(tup[s].symbol,
							tup[s].offset));

						/* Move mm0 into destination */
						Ctab();
						fprintf(Cout,
							"movq_r2m(mm0, %s);\n",
							spoolname (tup[s].reg =
							   tup[s].oreg));

						/* Restore mm0 */
						Ctab();
						fprintf(Cout,
							"movq_m2r(%s, mm0);\n",
							spoolname(REGSAVAIL-1));
					}
				} else if (optcpu & CPU_MAX) {
					if (tup[s].oreg < REGSAVAIL) {
						/* Didn't spill */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						fprintf(Cout,
						"movq_m2r(%s, r%d);\n",
						symname(tup[s].symbol,
							tup[s].offset),
						(tup[s].reg = tup[s].oreg));
					} else {
						/* Destination spilled */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						/* Save r0 */
						fprintf(Cout,
							"movq_r2m(r0, %s);\n",
							spoolname(REGSAVAIL-1));
						/* Move source into r0 */
						Ctab();
						fprintf(Cout,
						"movq_m2r(%s, r0);\n",
						symname(tup[s].symbol,
							tup[s].offset));

						/* Move r0 into destination */
						Ctab();
						fprintf(Cout,
							"movq_r2m(r0, %s);\n",
							spoolname (tup[s].reg =
							   tup[s].oreg));

						/* Restore r0 */
						Ctab();
						fprintf(Cout,
							"movq_m2r(%s, r0);\n",
							spoolname(REGSAVAIL-1));
					}
				} else if (optcpu == GenericIA32) {
					#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
					#endif
					fprintf(Cout,
					"movl_m2r(%s, %s);\n",
					symname(tup[s].symbol, tup[s].offset),
					ia32regname(tup[s].reg = tup[s].oreg));
				}
				break;
			case STORE:
				Ctab();
				if (optcpu & CPU_MMX) {
					if (tup[tup[s].arg[0]].reg<REGSAVAIL) {
						/* Didn't spill */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						fprintf(Cout,
						"movq_r2m(mm%d, %s);\n",
						tup[ tup[s].arg[0] ].reg,
						symname(tup[s].symbol,
							tup[s].offset));
					} else {
						/* Source spilled */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						/* Save mm0 */
						fprintf(Cout,
							"movq_r2m(mm0, %s);\n",
							spoolname(REGSAVAIL-1));
						/* Move source into mm0 */
						Ctab();
						fprintf(Cout,
						    "movq_m2r(%s, mm0);\n",
						    spoolname(
						       tup[tup[s].arg[0]].reg));

						/* Move mm0 into destination */
						Ctab();
						fprintf(Cout,
							"movq_r2m(mm0, %s);\n",
							symname(tup[s].symbol,
								tup[s].offset));

						/* Restore mm0 */
						Ctab();
						fprintf(Cout,
							"movq_m2r(%s, mm0);\n",
							spoolname(REGSAVAIL-1));
					}
				} else if (optcpu & CPU_MAX) {
					if (tup[tup[s].arg[0]].reg<REGSAVAIL) {
						/* Didn't spill */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						fprintf(Cout,
						"movq_r2m(r%d, %s);\n",
						tup[ tup[s].arg[0] ].reg,
						symname(tup[s].symbol,
							tup[s].offset));
					} else {
						/* Source spilled */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						/* Save r0 */
						fprintf(Cout,
							"movq_r2m(r0, %s);\n",
							spoolname(REGSAVAIL-1));
						/* Move source into r0 */
						Ctab();
						fprintf(Cout,
						    "movq_m2r(%s, r0);\n",
						    spoolname(
						       tup[tup[s].arg[0]].reg));

						/* Move r0 into destination */
						Ctab();
						fprintf(Cout,
							"movq_r2m(r0, %s);\n",
							symname(tup[s].symbol,
								tup[s].offset));

						/* Restore r0 */
						Ctab();
						fprintf(Cout,
							"movq_m2r(%s, r0);\n",
							spoolname(REGSAVAIL-1));
					}
				} else if (optcpu == GenericIA32) {
					#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
					#endif
					fprintf(Cout,
					"movl_r2m(%s, %s);\n",
					ia32regname(tup[ tup[s].arg[0] ].reg),
					symname(tup[s].symbol, tup[s].offset));
				}
				break;
			case STORER:
				Ctab();
				if (optcpu & CPU_MMX) {
					if ((tup[tup[s].arg[0]].reg<REGSAVAIL)
					 && (tup[tup[s].arg[1]].reg<REGSAVAIL))
					{
						/* Didn't spill */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						fprintf(Cout,
						"movq_r2x(mm%d, mm%d);\n",
						tup[ tup[s].arg[1] ].reg,
						tup[ tup[s].arg[0] ].reg);
					} else if
					   ((tup[tup[s].arg[0]].reg>=REGSAVAIL)
					 && (tup[tup[s].arg[1]].reg<REGSAVAIL))
					{
						/* EA spilled */
						int savereg =
						  (tup[tup[s].arg[1]].reg==0)?
						  1:0;

						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						/* Save savereg */
						fprintf(Cout,
							"movq_r2m(mm%d, %s);\n",
							savereg,
							spoolname(REGSAVAIL-1));

						/* Move EA into savereg */
						Ctab();
						fprintf(Cout,
						    "movq_m2r(%s, mm%d);\n",
						    spoolname(
						       tup[tup[s].arg[0]].reg),
						    savereg);

						/* Move data (arg1) to
						   (savereg) */
						Ctab();
						fprintf(Cout,
						    "movq_r2x(mm%d, mm%d);\n",
						    tup[tup[s].arg[1]].reg,
						    savereg);

						/* Restore savereg */
						Ctab();
						fprintf(Cout,
							"movq_m2r(%s, mm%d);\n",
							spoolname(REGSAVAIL-1),
							savereg);
					} else if
					   ((tup[tup[s].arg[0]].reg<REGSAVAIL)
					 && (tup[tup[s].arg[1]].reg>=REGSAVAIL))
					{
						/* Data spilled */
						/* Use the m2x form */
						fprintf(Cout,
						    "movq_m2x(%s, mm%d);\n",
						    spoolname(
							tup[tup[s].arg[1]].reg),
						    tup[ tup[s].arg[0] ].reg);
					} else if
					   ((tup[tup[s].arg[0]].reg>=REGSAVAIL)
					 && (tup[tup[s].arg[1]].reg>=REGSAVAIL))
					{
						/* Both spilled */
						/* Spill EA then use m2x form */

						int savereg =
						  (tup[tup[s].arg[1]].reg==0)?
						  1:0;

						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						/* Save savereg */
						fprintf(Cout,
							"movq_r2m(mm%d, %s);\n",
							savereg,
							spoolname(REGSAVAIL-1));

						/* Move EA into savereg */
						Ctab();
						fprintf(Cout,
						    "movq_m2r(%s, mm%d);\n",
						    spoolname(
						       tup[tup[s].arg[0]].reg),
						    savereg);

						/* Move data (arg1) to
						   (savereg) */
						Ctab();
						fprintf(Cout,
						    "movq_m2x(%s, mm%d);\n",
						    spoolname(
							tup[tup[s].arg[1]].reg),
						    savereg);

						/* Restore savereg */
						Ctab();
						fprintf(Cout,
							"movq_m2r(%s, mm%d);\n",
							spoolname(REGSAVAIL-1),
							savereg);
					}
				} else if (optcpu & CPU_MAX) {
					if (tup[tup[s].arg[0]].reg<REGSAVAIL) {
						/* Didn't spill */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						fprintf(Cout,
						"movqr_r2m(r%d, %s);\n",
						tup[ tup[s].arg[0] ].reg,
						symname(tup[s].symbol,
							tup[s].offset));
					} else {
						/* Source spilled */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						/* Save r0 */
						fprintf(Cout,
							"movqr_r2m(r0, %s);\n",
							spoolname(REGSAVAIL-1));
						/* Move source into r0 */
						Ctab();
						fprintf(Cout,
						    "movqr_m2r(%s, r0);\n",
						    spoolname(
						       tup[tup[s].arg[0]].reg));

						/* Move r0 into destination */
						Ctab();
						fprintf(Cout,
							"movqr_r2m(r0, %s);\n",
							symname(tup[s].symbol,
								tup[s].offset));

						/* Restore r0 */
						Ctab();
						fprintf(Cout,
							"movqr_m2r(%s, r0);\n",
							spoolname(REGSAVAIL-1));
					}
				} else if (optcpu == GenericIA32) {
					#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
					#endif
					fprintf(Cout,
					  "movl_r2x(%s, ",
					  ia32regname(tup[tup[s].arg[1]].reg));
					fprintf(Cout,
					  "%s);\n",
					  ia32regname(tup[tup[s].arg[0]].reg));
				}
				break;
			case NUM:
				cpoolenter(tup[s].immed);
				Ctab();
				if (optcpu & CPU_MMX) {
					if (tup[s].oreg < REGSAVAIL) {
						/* Didn't spill */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						fprintf(Cout,
						"movq_m2r(%s, mm%d);\n",
						cpoolname(tup[s].immed),
						(tup[s].reg = tup[s].oreg));
					} else {
						/* Destination spilled */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						/* Save mm0 */
						fprintf(Cout,
							"movq_r2m(mm0, %s);\n",
							spoolname(REGSAVAIL-1));
						/* Move source into mm0 */
						Ctab();
						fprintf(Cout,
						"movq_m2r(%s, mm0);\n",
						cpoolname(tup[s].immed));

						/* Move mm0 into destination */
						Ctab();
						fprintf(Cout,
							"movq_r2m(mm0, %s);\n",
							spoolname (tup[s].reg =
							   tup[s].oreg));
						/* Restore mm0 */
						Ctab();
						fprintf(Cout,
							"movq_m2r(%s, mm0);\n",
							spoolname(REGSAVAIL-1));
					}
				} else if (optcpu & CPU_MAX) {
					if (tup[s].oreg < REGSAVAIL) {
						/* Didn't spill */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						fprintf(Cout,
						"movq_m2r(%s, r%d);\n",
						cpoolname(tup[s].immed),
						(tup[s].reg = tup[s].oreg));
					} else {
						/* Destination spilled */
						#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
						#endif
						/* Save r0 */
						fprintf(Cout,
							"movq_r2m(r0, %s);\n",
							spoolname(REGSAVAIL-1));
						/* Move source into r0 */
						Ctab();
						fprintf(Cout,
						"movq_m2r(%s, r0);\n",
						cpoolname(tup[s].immed));

						/* Move r0 into destination */
						Ctab();
						fprintf(Cout,
							"movq_r2m(r0, %s);\n",
							spoolname (tup[s].reg =
							   tup[s].oreg));
						/* Restore r0 */
						Ctab();
						fprintf(Cout,
							"movq_m2r(%s, r0);\n",
							spoolname(REGSAVAIL-1));
					}
				} else if (optcpu == GenericIA32) {
					#ifdef DEBUG_CODEGEN
						fprintf(Cout,
							"/* %d */\n",
							__LINE__);
					#endif
					fprintf(Cout,
					"movl_m2r(%s, %s);\n",
					cpoolname(tup[s].immed),
					ia32regname(tup[s].reg = tup[s].oreg));
				}
				break;
			default:
				/* restore original register */
				tup[s].reg = tup[s].oreg;

				if (optcpu & CPU_MMX) {
					mmxop(s);
				} else if (optcpu & CPU_MAX) {
					maxop(s);
				} else if (optcpu == GenericIA32) {
					ia32op(s);
				} else {
					bug ("Unknown target in "
					     "output_schedule()");
					mmxop(s);
				}
			}
		}
}

static void
sched_bb()
{
	register int i;

#ifdef NEWSTUFF
	if (sched_simple() == NOTFOUND) {
		info(0, "basic block register allocator/scheduler failed");
		fflush(stderr);
	} else if (opttime > 0) {
#endif
		/* Search for a schedule using no spills... */
		if (sched_search(0) == NOTFOUND) {
			/* ... if none found, reset reference counts, then
			   look for a schedule allowing spills */
			refcounts();
			sched_search(MAXSPILLS);
		}
#ifdef NEWSTUFF
	}
#endif

	if (sbestval == NOTFOUND) {
		bug("basic block register allocator/scheduler failed");
	} else if (schedsp > 0) {
		/* Copy best schedule to sched[] */
		for (i=0; i<schedsp; ++i) {
			sched[i] = sbest[i];
		}

		/* Performance stats */
		if (optperf && schedsp) {
			register double d =
				(fieldcount / (0.5 * (schedsp + sbestval)));

			Ctab();
			fprintf(Cout,
				"/* block cost estimate is %d%s clocks, "
				"%1.1fx speedup */\n",
				((schedsp + sbestval) / 2),
				(((schedsp + sbestval) & 1) ? ".5" : ""),
				d);

			clocktotal += (schedsp + sbestval);
			fieldtotal += fieldcount;
		}

		/* Optionally dump list of scheduled tuples */
		#ifdef DEBUG_SCHED
			dump_sched(stderr, "- final version");
		#endif

		output_schedule();
	}
}

void
end_bb(void)
{
	register int i, j;

	#ifdef DEBUG_ENDBB
		info(0, "Start end_bb()");
	#endif

	#ifdef DEBUG_TUPLIST
		dump_tuples(stderr, "at start of end_bb()");
	#endif

	/* Don't schedule empty blocks (this should be checked elsewere) */
	if (tupsp == -1) return;

	/* First set reference counts */
	refcounts();

	/* Pull immediate values into the instructions...
	   Intel recommends that single-use constants be
	   pulled into the instructions that use them;
	   this loop effectively does that.
	*/
	for (i=0; i<tupsp; ++i) {
		tup[i].m2rmode = -1;
	}

#ifdef	NEVER
	for (i=0; i<tupsp; ++i) {
		if ((tup[i].refs > 0) &&
		    (tup[i].op != STORE) &&
		    (tup[i].op != STORER)) {
			register int j = tup[i].arg[1];

			if (j != -1) {
				if (((tup[j].op == NUM) ||
				     (tup[j].op == LOAD)) &&
				    (tup[j].refs == 1)) {
					tup[i].m2rmode = j;
					tup[j].refs = 0;
					tup[i].arg[1] = -1;
				}

				j = tup[i].arg[0];
				if ((j != -1) &&
				    (tup[i].m2rmode == -1) &&
				    (!ordered(tup[i].op)) &&
				    ((tup[ j ].op == NUM) ||
				     (tup[ j ].op == LOAD)) &&
				    (tup[ j ].refs == 1)) {
					/* swap args and absorb constant */
					tup[i].m2rmode = j;
					tup[j].refs = 0;
					tup[i].arg[0] = tup[i].arg[1];
					tup[i].arg[1] = -1;
				}
			}
		}
	}

	/* Set reference counts */
	refcounts();
#endif

	#ifdef DEBUG_PERF
		fprintf(stderr, "%s: tupsp=%d\n", funcname, tupsp);
		fflush(stderr);
	#endif

	/* Now do before and after fields */
	before_bb(0,0,1);	/* This just inits the static cache */
	for (i=0; i<tupsp; ++i) {
		#ifdef DEBUG_ENDBB
			fprintf(stderr, "\ni=%d\n  j=", i);
			fflush(stderr);
		#endif
		if (tup[i].refs > 0) {
			tup[i].before = 0;
			tup[i].after = 0;
			for (j=0; j<tupsp; ++j) if (i != j) {
				#ifdef DEBUG_ENDBB
					fprintf(stderr, "%d ", j);
					fflush(stderr);
				#endif
				if (before_bb(j, i, 0))
					++(tup[i].before);
				else if (before_bb(i, j, 0))
					++(tup[i].after);
			}
		}
	}

	#ifdef DEBUG_ENDBB
		info(0, "end_bb(): calling sched_bb()");
	#endif

	/* Schedule the basic block */
	sched_bb();

	#ifdef DEBUG_ENDBB
		info(0, "End of end_bb()");
	#endif
}


inline int
sched_pos(register int i)
{
	/* Find the position in the schedule of tuple i */

	register int j;

	for (j=0; j<schedsp; ++j) {
		if (sched[j] == i) return(j);
	}

	{
		char buf[64];
		snprintf(buf, 64, "sched_pos could not find instruction %d "
				  "(this cannot happen?)", i);
		bug(buf);
	}
	return(-1);
}

inline int
sched_class(int op)
{
	/* Determines MMX internal resource usage for the purpose
	   of estimating conflicts, hence performance
	*/

	switch (op) {
	case AND:
	case ANDN:
	case OR:
	case XOR:
	case NOT:
	case ADD:
	case EQ:
	case EQ_C:
	case GE:
	case GT:
	case GT_C:
	case SUB:
	case NEG:
	case LEA:
	case STORE:
	case STORER:
		return(0);
	case AVG:
	case DIV:
	case MOD:
	case I2F:
	case F2I:
	case RCP:
	case RCP1:
	case RCP2:
		return(0);
	case MUL:
	case MULH:
		return(1);
	case SHL:
	case SHR:
	case PACK:
	case UNPACKL:
	case UNPACKH:
		return(2);
	case LOAD:
	case LOADR:
	case NUM:
		return(4);
	case MIN:
	case MAX:
		return(8);
	case MIXEVEN:
	case MIXODD:
	case PERM:
		info("sched_class() doesn't correctly handle PA-RISC MAX");
		return(2);
	default:
		{
			char buf[64];
			snprintf(buf,
				64,
				"sched_class op=%s (this cannot happen?)",
				opname(op));
			bug(buf);
		}
		return(0);
	}
}

inline int
sched_ok(register int i,
register int pos)
{
	/* Returns 1 if it is okay to schedule tuple i in position pos?
	   Returns 0 otherwise.
	*/

	register int j, k;

	/* Do the quick before and after checks */
	if (pos < tup[ sched[i] ].before) return(0);
	if (pos >= (schedsp - tup[ sched[i] ].after)) return(0);

	/* Are the immediate predecessors scheduled before this? */
	if (((k = tup[ sched[i] ].arg[0]) != -1) &&
	    (sched_pos(k) >= pos)) return(0);
	if (((k = tup[ sched[i] ].arg[1]) != -1) &&
	    (sched_pos(k) >= pos)) return(0);

	/* If this is a store, is anti-dependence ok? */
	if (tup[ sched[i] ].op == STORE) {
		/* no load of the same object should follow this */
		for (j=i; j<schedsp; ++j) {
			/* Is this an explicit load? */
			if ((tup[ sched[j] ].op == LOAD) &&
			    (tup[ sched[j] ].symbol ==
			     tup[ sched[i] ].symbol) &&
			    (tup[ sched[j] ].offset ==
			     tup[ sched[i] ].offset)) {
				/* Load of same object */
				return(0);
			}

			/* Is this an m2r load? */
			if ((tup[ sched[j] ].m2rmode != -1) &&
			    (tup[ tup[ sched[j] ].m2rmode ].symbol ==
			     tup[ sched[i] ].symbol) &&
			    (tup[ tup[ sched[j] ].m2rmode ].offset ==
			     tup[ sched[i] ].offset)) {
				/* Implicit load of same object */
				return(0);
			}
		}
	}

	/* Don't schedule this tuple ahead of any previous tuple with which
	   it may have a conflict.  Effectively, this means that no lower-
	   numbered conflicting tuple should follow this in the schedule. */
	for (j=i+1; j<schedsp; ++j) {
		if ((sched[i]>sched[j]) &&
		    (possible_conflict(sched[i],sched[j])))
			return(0);
	}

	/* Must be ok...? */
	return(1);
}

inline int
simcost(register int s)
{
	/* Return estimated cost of instructions simulated in
	   target- specific header file */
	register int val = 0;

	switch (tup[s].op) {
	case DIV:
	case MOD:
		/* Approximate cost as per serial code estimate */
		val = 2 + 3 * (64 / bitsperfield(tup[s].type.bits));
		break;
	case MUL:
		if (bitsperfield(tup[s].type.bits) == 32) {
			/* This is also faked by serial code */
			val = 2 + 3 * 2;
		}
		break;
	}

	if (((optcpu & CPU_2PIPE) == 0) && (val & 1)) ++val;
	return(val);
}

inline int
sched_cost(register int pos)
{
	/* Compute approximate cost of this placement */
	register int val = 0;
	register int val0 = 0;
	register int val1 = 0;
	register int s = sched[pos];

	/* Is this an expensive simulated op? */
	if ((val = simcost(s)) > 0) {
		/* Simulated instructions are S L O W...
		   other delays are negligible
		*/
		return(val);
	}

	/* Two same-class ops in a row? */
	if ((pos > 0) &&
	    ((sched_class(tup[s].op) &
	      sched_class(tup[ sched[pos - 1] ].op)))) {
		/* Expected cost of 0-1/2 clock */
		val += ((optcpu & CPU_2PIPE) ? 1 : 0);
	}

	/* Cyrix MAX instruction takes 2 clocks */
	if ((optcpu & CPU_XMMX) &&
	    (tup[s].op == MAX)) {
		val += 4;
	}

	/* Where are the operands coming from? */
	if (tup[s].arg[1] != -1) {
		register int i = sched_pos(tup[s].arg[1]);
		register int dist = 2;

		switch (tup[sched[i]].op) {
		case LOAD:
		case NUM:
			dist = 4;
			break;
		case MUL:
			dist = ((optcpu == K6_2) ? 4 : 6);
		}

		val1 = (dist / (pos - i));
		if (((optcpu & CPU_2PIPE) == 0) && (val1 & 1)) ++val1;
	}
	if (tup[s].arg[0] != -1) {
		register int i = sched_pos(tup[s].arg[0]);
		register int dist = 2;

		switch (tup[sched[i]].op) {
		case LOAD:
		case NUM:
			dist = 4;
			break;
		case MUL:
			dist = ((optcpu == K6_2) ? 4 : 6);
		}

		val0 += (dist / (pos - i));
		if (((optcpu & CPU_2PIPE) == 0) && (val0 & 1)) ++val0;
	}
	if (val0 > val1) {
		val += val0;
	} else {
		val += val1;
	}

	/* Also have to give a penalty for inserted movq */
	if ((tup[s].antidep != -1) &&
	    (tup[s].op != STORE) &&
	    (tup[s].op != STORER)) {
		/* 1/2-1 clock penalty */
		val += ((optcpu & CPU_2PIPE) ? 1 : 2);
	}

	return(val);
}

inline int
sched_eval(void)
{
	/* If the candidate schedule is better than the current best,
	   save it as the current best.
	   Return 1 if the absolute best schedule has been found, or
	    if the timelimit has been reached.
	*/

	register int i;

	if ((sbestval == NOTFOUND) ||
	    (schedval < sbestval)) {
		for (i=0; i<schedsp; ++i) {
			sbest[i] = sched[i];
		}
		sbestval = schedval;
#ifdef NOTDEFD
/* I don't think this should be done... */
		schedtim = (time(0) + opttime);
#endif
	}

	/* Timeout if needed... */
	if (time(0) >= schedtim) {
		return(1);
	}

	/* Is this as good as it gets? */
	return(sbestval == 0);
}

